#include <ttg.h>

using Key2 = std::pair<int, int>;
namespace std
{
    std::ostream &operator<<(std::ostream &os,
                             const Key2 &key)
    {
        os << "{" << std::get<0>(key) << ", "
           << std::get<1>(key) << "}";
        return os;
    }
} // namespace std

int main(int argc, char **argv)
{
    ttg::initialize(argc, argv, -1);

    ttg::Edge<Key2, double> A_B("A->B");
    ttg::Edge<int, double> B_C0("B->C0");
    ttg::Edge<int, double> B_C1("B->C1");
    ttg::Edge<int, double> C_A("C->A");

    double threshold = 100.0;

    auto wa = ttg::make_tt(
        [](const int &it,
           const double &a)
        {
            ttg::send<0>(Key2{it, 0}, a);
            ttg::send<0>(Key2{it, 1}, a+1.0); 
        },
        ttg::edges(C_A), ttg::edges(A_B), "A",
        {"from C"}, {"to B"});

    auto wb = ttg::make_tt(
        [](const Key2 &k,
           const double &a)
        {
            if(0 == std::get<1>(k)) ttg::send<0>(std::get<0>(k), 2.0*a);
            if(1 == std::get<1>(k)) ttg::send<1>(std::get<0>(k), 2.0*a); 
        },
        ttg::edges(A_B), ttg::edges(B_C0, B_C1),
        "B", {"from A"},
        {"to 1st input of C", "to 2nd input of C"});

    auto wc = ttg::make_tt(
        [=](const int &it,
            const double &a,
            const double &b)
        {
            if(a+b < threshold) ttg::send<0>(it+1, a+b);
            else ttg::print("Result ", a + b ); 
        },
        ttg::edges(B_C0, B_C1), ttg::edges(C_A),
        "C", {"From B", "From B"}, {"To A"});

    ttg::make_graph_executable(wa);

    if (wa->get_world().rank() == 0)
        wa->invoke(0, 0.0);

    ttg::execute();
    ttg::fence(wa->get_world());

    ttg::finalize();
    return EXIT_SUCCESS;
}
